import sqlalchemy
import zodchy
import collections.abc

from ..contracts import Clause, ClauseExpression
from .joins import JoinsAssembler
from .orders import OrdersAssembler
from .filters import FilterAssembler
from .slices import SlicesAssembler


class QueryAssembler:
    def __init__(self, query: sqlalchemy.Select):
        self._query = query

    def __call__(self, *clauses: Clause | ClauseExpression | zodchy.codex.query.SliceBit) -> sqlalchemy.Select:
        filters, orders, slices = self._separate(clauses)
        filter_expression = self._build_expression(filters)
        self._query = JoinsAssembler(self._query)(filter_expression)
        if (filters := FilterAssembler()(filter_expression)) is not None:
            self._query = self._query.where(filters)
        self._query = OrdersAssembler(self._query)(*orders)
        self._query = SlicesAssembler(self._query)(*slices)
        return self._query

    @staticmethod
    def _separate(clauses: collections.abc.Iterable[Clause | ClauseExpression]):
        filters = []
        orders = []
        slices = []
        for clause in clauses:
            if isinstance(clause, Clause):
                if zodchy.codex.query.FilterBit in clause.operation.__class__.__mro__:
                    filters.append(clause)
                elif zodchy.codex.query.OrderBit in clause.operation.__class__.__mro__:
                    orders.append(clause)
                else:
                    raise ValueError(f'Expected a filter, order or slice clause, got {clause.operation}')
            elif isinstance(clause, ClauseExpression):
                filters.append(clause)
            elif isinstance(clause, zodchy.codex.query.SliceBit):
                slices.append(clause)
        return filters, orders, slices

    @staticmethod
    def _build_expression(clauses: collections.abc.Iterable[Clause | ClauseExpression]) -> ClauseExpression | None:
        expression = None
        for clause in clauses:
            if isinstance(clause, Clause):
                if expression is None:
                    expression = ClauseExpression(clause)
                else:
                    expression = expression & clause
            elif isinstance(clause, ClauseExpression):
                if expression is None:
                    expression = clause
                else:
                    expression = expression & clause
        return expression



import sqlalchemy
import zodchy

from ..contracts import Clause, ClauseExpression


class OrdersAssembler:
    def __init__(self, query: sqlalchemy.Select):
        self._query = query

    def __call__(self, *clauses: Clause) -> sqlalchemy.Select:
        for clause in clauses:
            if isinstance(clause, Clause) and isinstance(clause.operation, zodchy.operators.OrderBit):
                self._assemble(clause)
        return self._query

    def _assemble(self, clause: Clause):
        if isinstance(clause.operation, zodchy.operators.ASC):
            self._query = self._query.order_by(clause.column.asc())
        elif isinstance(clause.operation, zodchy.operators.DESC):
            self._query = self._query.order_by(clause.column.desc())



import typing

import zodchy
import sqlalchemy
from sqlalchemy import BinaryExpression, Table

from ..contracts import Clause, ClauseExpression


class JoinsAssembler:
    def __init__(self, query: sqlalchemy.Select):
        self._query = query
        self._joins_digest: set[str] = set()
        self._tables: set[str] = set(t.name for t in query.columns_clause_froms)
        self._foreign_keys: dict[str, sqlalchemy.ForeignKey] = {}
        self._prepare()

    def __call__(self, expression: ClauseExpression) -> sqlalchemy.Select:
        for clause in expression or ():
            if isinstance(clause, Clause) and isinstance(clause.operation, zodchy.operators.FilterBit):
                self._build_link(clause)
        return self._query

    def _build_link(self, clause: Clause):
        column = clause.column
        table_name = column.table.name
        if table_name in self._tables:
            return

        for condition in clause.conditions:
            if isinstance(condition, Table):
                for fk in condition.foreign_keys:
                    if fk.column.table.name in self._tables:
                        condition = fk.column == fk.parent
            if isinstance(condition, BinaryExpression):
                if (joined_table := self._register_join(condition)) is not None:
                    self._query = self._query.join(
                        joined_table,
                        condition,
                        isouter=True
                    )
                    self._register_foreign_keys(joined_table)



        for fk_table_name, fk in self._foreign_keys.items():
            if fk_table_name == table_name:
                if self._register_join(
                    typing.cast(
                        sqlalchemy.BinaryExpression,
                        fk.column == fk.parent
                    )
                ) is not None:
                    self._query = self._query.join(column.table, fk.column == fk.parent, isouter=True)
                break

    def _prepare(self):
        for join in self._query._setup_joins or ():  # todo: should think how to fix this hack
            self._register_join(join[1])

        for entity in self._query.froms:
            if isinstance(entity, sqlalchemy.Table):
                self._register_foreign_keys(entity)

    def _register_join(self, expression: sqlalchemy.BinaryExpression) -> sqlalchemy.Table | None:
        result = None
        if (digest := self._assemble_join_digest(expression)) not in self._joins_digest:
            self._joins_digest.add(digest)
            if expression.left.table.name not in self._tables:
                self._tables.add(expression.left.table.name)
                result = expression.left.table
            if expression.right.table.name not in self._tables:
                self._tables.add(expression.right.table.name)
                result = expression.right.table
        return result

    def _register_foreign_keys(self, table: sqlalchemy.Table):
        for foreign_key in table.foreign_keys:
            table_name = foreign_key.column.table.name
            if table_name not in self._foreign_keys:
                self._foreign_keys[table_name] = foreign_key

    @staticmethod
    def _assemble_join_digest(expression: sqlalchemy.BinaryExpression) -> str | None:
        names = []
        for element in expression.left, expression.right:
            if isinstance(element, sqlalchemy.Column):
                names.append(element.table.name)
            else:
                return
        return "::".join(sorted(names))



import typing
import collections.abc

import sqlalchemy
import zodchy

AND = typing.NewType('AND', object)
OR = typing.NewType('OR', object)


class Clause:
    def __init__(
        self,
        column: sqlalchemy.Column,
        operation: zodchy.codex.query.ClauseBit,
        *conditions: sqlalchemy.BinaryExpression | sqlalchemy.Table
    ):
        self.column = column
        self.operation = operation
        self.conditions = conditions

    def __and__(self, other: typing.Self | 'ClauseExpression'):
        if self.operation is None:
            return other
        self_obj = Clause(self.column, self.operation, *self.conditions)
        if isinstance(other, Clause):
            if other.operation is None:
                return self
            else:
                return ClauseExpression(other, self_obj, AND)
        elif isinstance(other, ClauseExpression):
            return ClauseExpression(self_obj, *other, AND)

    def __or__(self, other: typing.Self | 'ClauseExpression'):
        if self.operation is None:
            return other
        self_obj = Clause(self.column, self.operation, *self.conditions)
        if isinstance(other, Clause):
            if other.operation is None:
                return self
            else:
                return ClauseExpression(other, self_obj, OR)
        elif isinstance(other, ClauseExpression):
            return ClauseExpression(self_obj, *other, OR)

    def clone(self) -> typing.Self:
        return Clause(self.column, self.operation, *self.conditions)

    def dump(self) -> dict:
        return dict(
            column=self.column,
            operation=self.operation,
            conditions=self.conditions
        )


class ClauseExpression:
    def __init__(self, *clauses: Clause | AND | OR):
        self._clauses = list(self._assure_filter_clause(clauses))

    def __iter__(self):
        yield from self._clauses

    def __and__(self, other: typing.Self | Clause):
        if isinstance(other, Clause):
            if other.operation is None:
                return self
            else:
                return ClauseExpression(other, *self._clauses, AND)
        elif isinstance(other, ClauseExpression):
            return ClauseExpression(*self._clauses, *other, AND)

    def __or__(self, other: typing.Self | Clause):
        if isinstance(other, Clause):
            if other.operation is None:
                return self
            else:
                return ClauseExpression(other, *self._clauses, OR)
        elif isinstance(other, ClauseExpression):
            return ClauseExpression(*self._clauses, *other, OR)

    @staticmethod
    def _assure_filter_clause(
        clauses: collections.abc.Iterable[Clause]
    ) -> collections.abc.Generator[None, None, Clause]:
        for c in clauses:
            if c is AND or c is OR or zodchy.codex.query.FilterBit in c.operation.__class__.__mro__:
                yield c
            else:
                raise ValueError(f'Expected a filter clause, got {c.operation}')



from .joins import JoinsAssembler
from .queries import QueryAssembler
from .filters import FilterAssembler
from .orders import OrdersAssembler
from .slices import SlicesAssembler
from .mutations import MutationAssembler


import collections.abc

import sqlalchemy
import zodchy

from .. import contracts
from .filters import FilterAssembler

DataRow = collections.abc.Mapping


class MutationAssembler:
    def __init__(self, table: sqlalchemy.Table):
        self._table = table
        self._filter_assembler = FilterAssembler()

    def __call__(self, *elements: DataRow | contracts.Clause | contracts.ClauseExpression):
        data, filters = self._separate(elements)
        if data and filters:
            return self._update(data, filters)
        elif len(data) > 0 and len(filters) == 0:
            return self._insert(data)
        elif len(data) == 0 and len(filters) > 0:
            return self._delete(filters)

    def _update(self, data: list[DataRow], filters: list[contracts.Clause | contracts.ClauseExpression]):
        if len(data) > 1:
            raise ValueError(f'Expected only one data item, got {len(data)}')
        if len(filters) == 0:
            raise ValueError(f'Expected at least one filter, got {len(filters)}')
        return sqlalchemy.update(self._table).values(**data[0]).where(self._filter_assembler(*filters))

    def _insert(self, data: list[DataRow]):
        return sqlalchemy.insert(self._table).values(data)

    def _delete(self, filters: list[contracts.Clause | contracts.ClauseExpression]):
        return sqlalchemy.delete(self._table).where(self._filter_assembler(*filters))

    @staticmethod
    def _separate(
        elements: collections.abc.Iterable[DataRow | contracts.Clause | contracts.ClauseExpression]
    ):
        data = []
        filters = []

        for element in elements:
            if isinstance(element, contracts.Clause):
                if zodchy.codex.query.FilterBit in element.operation.__class__.__mro__:
                    filters.append(element)
                else:
                    raise ValueError(f'Expected a filter, got {element.operation}')
            elif isinstance(element, contracts.ClauseExpression):
                filters.append(element)
            elif isinstance(element, DataRow):
                data.append(element)

        return data, filters



import sqlalchemy
import zodchy


class SlicesAssembler:
    def __init__(self, query: sqlalchemy.Select):
        self._query = query

    def __call__(self, *operations: zodchy.operators.SliceBit) -> sqlalchemy.Select:
        for operation in operations:
            self._assemble(operation)
        return self._query

    def _assemble(self, operation: zodchy.operators.SliceBit):
        if isinstance(operation, zodchy.operators.Limit):
            self._query = self._query.limit(operation.value)
        elif isinstance(operation, zodchy.operators.Offset):
            self._query = self._query.offset(operation.value)



import typing
import uuid
import warnings
from functools import singledispatch

from sqlalchemy import Row


def to_dict(data: Row) -> dict[str, typing.Any]:
    result = {}
    for k, v in data._mapping.items():
        result[k] = field_serializer(v)
    return result


@singledispatch
def field_serializer(value):
    return value


try:
    import asyncpg.pgproto.pgproto


    @field_serializer.register
    def _(value: asyncpg.pgproto.pgproto.UUID) -> uuid.UUID:
        return uuid.UUID(bytes=value.bytes)
except Exception as e:
    warnings.warn(f'asyncpg not installed: {e}')



import typing
import operator
from collections import deque

import sqlalchemy
import zodchy

from ..contracts import Clause, ClauseExpression, AND, OR


class FilterAssembler:
    def __call__(self, clause: Clause | ClauseExpression):
        return self._assemble(clause)

    def _assemble(self, clause: Clause | ClauseExpression) -> sqlalchemy.ColumnElement:
        expression = ClauseExpression(clause) if isinstance(clause, Clause) else clause
        _op_map = {
            AND: sqlalchemy.and_,
            OR: sqlalchemy.or_
        }
        operations_stack: deque = deque(
            list(
                e for e in expression or ()
                if e is AND or e is OR or zodchy.operators.FilterBit in e.operation.__class__.__mro__
            )
        )
        if operations_stack:
            buffer = deque()
            while operations_stack:
                element = operations_stack.popleft()
                if element is AND or element is OR:
                    operands = []
                    i = 0
                    while buffer and i < 2:
                        operands.append(buffer.popleft())
                        i += 1
                    if operands:
                        buffer.appendleft(
                            _op_map[element](*operands) if len(operands) > 1 else operands[0]
                        )
                else:
                    buffer.appendleft(self._assemble_element(element))
            return buffer.popleft()

    def _assemble_element(
        self,
        element: sqlalchemy.ColumnElement | Clause,
    ):
        if isinstance(element, Clause):
            element = self._operations[type(element.operation)](element)
        return element

    @property
    def _operations(self):
        return {
            zodchy.operators.EQ: self._simple_clause(operator.eq),
            zodchy.operators.NE: self._simple_clause(operator.ne),
            zodchy.operators.LE: self._simple_clause(operator.le),
            zodchy.operators.LT: self._simple_clause(operator.lt),
            zodchy.operators.GE: self._simple_clause(operator.ge),
            zodchy.operators.GT: self._simple_clause(operator.gt),
            zodchy.operators.IS: lambda v: v.column.is_(v.operation.value),
            zodchy.operators.LIKE: self._like_clause,
            zodchy.operators.NOT: self._not_clause,
            zodchy.operators.SET: self._set_clause,
            zodchy.operators.RANGE: self._range_clause,
        }

    def _not_clause(self, clause: Clause) -> typing.Any:
        if isinstance(clause.operation, zodchy.operators.IS):
            return clause.column.isnot(clause.operation.value)
        elif isinstance(clause.operation, zodchy.operators.EQ):
            return operator.ne(clause.column, clause.operation.value)
        elif isinstance(clause.operation, zodchy.operators.LIKE):
            return self._like_clause(
                clause.clone(),
                inversion=True
            )
        elif isinstance(clause.operation, zodchy.operators.SET):
            return self._set_clause(
                clause.clone(),
                inversion=True
            )
        else:
            return sqlalchemy.not_(self._assemble(clause)),

    @staticmethod
    def _simple_clause(op):
        return lambda v: op(v.column, v.operation.value)

    def _logic_clause(self, op):
        return lambda v: op(self._assemble(u) for u in v)

    @staticmethod
    def _like_clause(clause: Clause, inversion: bool = False):
        column = clause.column
        operation = clause.operation
        value = f'%{operation.value}%'
        if hasattr(operation, 'case_sensitive') and operation.case_sensitive:
            return column.notlike(value) if inversion else column.like(value)
        else:
            return column.notilike(value) if inversion else column.ilike(value)

    @staticmethod
    def _set_clause(clause: Clause, inversion: bool = False):
        column = clause.column
        value = list(clause.operation.value)
        if inversion:
            return column.notin_(value)
        else:
            return column.in_(value)

    def _range_clause(self, clause: Clause):
        params = [
            Clause(clause.column, condition, *clause.conditions)
            for condition in clause.operation.value
            if condition is not None
        ]
        if len(params) > 1:
            return self._logic_clause(sqlalchemy.and_)(params)
        elif len(params) == 1:
            return params[0]
